# Copyright 2020 Pasqal Cloud Services development team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from sys import version_info
from typing import Any, Dict, List, Optional

if version_info[:2] >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict


class JobResult(TypedDict):
    raw: List[str]
    counter: Dict[str, Any]


class PaginationType(TypedDict):
    total: int
    offset: int
    end: int


class JSendPayload(TypedDict):
    status: str
    code: int
    message: str
    data: Any
    pagination: Optional[PaginationType]
